﻿/**
 * ARTool 2
 * 
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *      
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2010-2021, Jacob O. Wobbrock, Ph.D.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock OR The University
 * of Washington BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;
using WobbrockLib.Extensions;

namespace ARTool
{
    public class AlignedRankTransformer
    {
		private const decimal SumEpsilon = 0.0001m;

		/// <summary>
		/// Creates an instance of a class capable of performing the Aligned Rank Transform.
		/// </summary>
		public AlignedRankTransformer()
        {
            // do nothing
        }

        /// <summary>
        /// Parses the *.csv long-format data table from the given full-path filename.
        /// </summary>
        /// <param name="file">A full path to the file given.</param>
        /// <param name="delimiter">The character delimiter used to separate columns.</param>
        /// <returns>A new data table for the source data, or null if there is an error.</returns>
        /// <remarks>If a row is read in with the wrong number of tokens (wrong number
        /// of columns), an exception is thrown.</remarks>
        public LongFormatDataTable ParseDataTable(string file, char delimiter)
        {
            LongFormatDataTable table = new LongFormatDataTable();
            StreamReader reader = new StreamReader(file, Encoding.ASCII);
            string line = reader.ReadLine(); // column names line
            List<string> tokens = Tokenize(line, delimiter); // column names
            for (int i = 0; i < tokens.Count; i++)
            {
                table.AddColumn(new LongFormatDataColumn(tokens[i])); // add column names
            }
            int rowNo = 0;
            while (!reader.EndOfStream)
            {
                tokens.Clear();
                rowNo++;
                line = reader.ReadLine();
                tokens = Tokenize(line, delimiter);
                if (tokens.Count != table.NumColumns)
                    throw new FormatException(String.Format("Number of columns ({0}) does not equal number of column headers ({1}) in table row {2}. ", tokens.Count, table.NumColumns, rowNo));
                for (int i = 0; i < tokens.Count; i++)
                    table[i].AddRow(tokens[i]);
            }
            reader.Close();
            return table;
        }

        /// <summary>
        /// Creates the new columns for the output data table. These will contain columns from the 
        /// original source table, as well as the aligned columns and ranked columns. The names for 
        /// the new columns are based on the factor and response column names.
        /// </summary>
        /// <param name="srcTable">The source data table from which we'll create the column names
        /// in the destination data table.</param>
        /// <returns>The data table with columns added for the outputted data. No actual
        /// output data values have been added to the columns at this point.</returns>
        /// <param name="plainColNames">If true, plain column names are used, which avoid spaces, 
        /// parentheses, hyphens, and asterisks, and instead use only letters and underscores for
        /// these things. Original column names, if they had such characters, are left unaffected.</param>
        public LongFormatDataTable CreateNewColumnNames(LongFormatDataTable srcTable, bool plainColNames)
        {
            LongFormatDataTable dstTable = new LongFormatDataTable();

            // re-add the factor names that exist in the data table. this means our original
            // response measure (Y), the former last column, will be incorrectly regarded as a 
            // factor because it will no longer be last, but this is okay because we are only 
            // reading from the source data table, not the destination table.
            for (int i = 0; i < srcTable.NumColumns; i++)
            {
                dstTable.AddColumn(new LongFormatDataColumn(srcTable[i].Name));
            }

            // add the column names for all the aligned main effects and interactions
            List<List<string>> factorPset = SetsEx.PowerSet<string>(srcTable.FactorNames); // factor powerset
            for (int i = 1; i < factorPset.Count; i++) // skip the empty set at [0]
            {
                string colname = String.Format(plainColNames ? "aligned_{0}_for_" : "aligned({0}) for ", srcTable.ResponseName);
                for (int j = 0; j < factorPset[i].Count; j++)
                {
                    colname += String.Format(plainColNames ? "{0}_x_" : "{0}*", factorPset[i][j]);
                }
                colname = plainColNames ? colname.Substring(0, colname.LastIndexOf("_x_")) : colname.TrimEnd(new char[1] { '*' });
                dstTable.AddColumn(new LongFormatDataColumn(colname));
            }
            // add the column names for all the averaged ranks main effects and interactions
            for (int i = 1; i < factorPset.Count; i++)
            {
                string colname = String.Format(plainColNames ? "ART_{0}_for_" : "ART({0}) for ", srcTable.ResponseName);
                for (int j = 0; j < factorPset[i].Count; j++)
                {
                    colname += String.Format(plainColNames ? "{0}_x_" : "{0}*", factorPset[i][j]);
                }
                colname = plainColNames ? colname.Substring(0, colname.LastIndexOf("_x_")) : colname.TrimEnd(new char[1] { '*' });
                dstTable.AddColumn(new LongFormatDataColumn(colname));
            }
            return dstTable;
        }

		/// <summary>
		/// Computes the aligned data and adds that data to the aligned columns in the destination
		/// data table.
		/// </summary>
		/// <param name="srcTable">The source data table from which the calculations are made.</param>
		/// <param name="dstTable">The destination data table into which the aligned data is put.</param>
		/// <remarks>This alignment procedure follows that of statistician J.J. Higgins, who worked out
		/// the mathematics for an N-way factorial and communicated them to me in a personal whitepaper. I
		/// translated his mathematics into the algorithm here.</remarks>
		public void ComputeAlignedColumns(LongFormatDataTable srcTable, ref LongFormatDataTable dstTable)
        {
            List<List<string>> factorPset = SetsEx.PowerSet<string>(srcTable.FactorNames); // factor powerset; ignore [0] (empty set)
            Debug.Assert(factorPset.Count - 1 == (int) Math.Pow(2.0, srcTable.NumFactors) - 1.0, "Incorrect count for number of new columns.");

            SetSizeComparer sComp = new SetSizeComparer();
            decimal grandMean = srcTable.GetGrandMean();

            // walk down each row of the data table, adjusting each point (e.g., Y -> adjY) appropriately
            // for each new main or interaction effect created for the aligned data table.
            for (int row = 1; row <= srcTable.NumRows; row++)
            {
                string S = srcTable.GetSubjectForRow(row); // get the level of Subject for this row
                List<string> Xs = srcTable.GetLevelsOfFactorsForRow(row); // get the levels of X's for this row
                decimal Y = srcTable.GetResponseForRow(row); // get Y response that we will adjust ("align")

                // retain the original columns in our new table
                dstTable[0].AddRow(S); // subject identifier S
                for (int i = 0; i < Xs.Count; i++)
                {
                    dstTable[i + 1].AddRow(Xs[i]); // add all factors' (X's) levels
                }
                dstTable[Xs.Count + 1].AddRow(Y); // unaligned Y response
                Debug.Assert(Xs.Count + 2 == srcTable.NumColumns, "Number of columns do not agree."); // number of kept columns (S + X's + Y)

                // compute the residual
                decimal cellMean = srcTable.GetMeanForLevelsOfFactors(srcTable.FactorNames, Xs);
                decimal residual = Y - cellMean;

                // levelsPset[i] corresponds to factor factorPset[i]
                List<List<string>> levelsPset = SetsEx.PowerSet<string>(Xs);

                // compute effect for each factor or interaction column
                for (int i = 1; i < factorPset.Count; i++)
                {
                    List<List<string>> efxPset = SetsEx.PowerSet<string>(factorPset[i]); // get powerset of efx being considered
                    efxPset.Sort(sComp); // sorts in ascending cardinality
                    List<List<string>> lvlPset = SetsEx.PowerSet<string>(levelsPset[i]); // get powerset of levels being considered
                    lvlPset.Sort(sComp); // sorts in ascending cardinality

                    decimal efx = 0m; // the effect of interest
                    decimal sign = +1m; // used to flip +1/-1

                    // this is the J.J. Higgins alignment approach
                    for (int j = efxPset.Count - 1; j > 0; j--) // go from highest ordinality to least for subeffects
                    {
                        efx += sign * srcTable.GetMeanForLevelsOfFactors(efxPset[j], lvlPset[j]);
                        if (efxPset[j].Count != efxPset[j - 1].Count)
                            sign = -sign; // flip sign when cardinality changes
                    }
                    efx += sign * grandMean;

                    decimal adjY = residual + efx; // aligned response
                    dstTable[Xs.Count + 1 + i].AddRow(adjY);
                }
            } // end of iteration down rows
        }

        /// <summary>
        /// Sums the aligned columns in the given table and ensures that they sum to near-zero.
        /// </summary>
        /// <param name="nFactors">The number of factots from which this destination table was created.</param>
        /// <param name="dstTable">The table in which to sum its aligned columns, which need to have been
        /// already created.</param>
        /// <returns>A list of the column names who are problematic, if any. Will not be null, but 
        /// may have zero elements.</returns>
        public List<string> SumAlignedColumns(int nFactors, LongFormatDataTable dstTable)
        {
            int startIdx = nFactors + 2; // nFactors + 2 b/c the src table has cols S, X1..Xn, Y
            int count = (int) Math.Pow(2.0, nFactors) - 1;
            List<string> problems = new List<string>(count);

            for (int i = startIdx; i < startIdx + count; i++)
            {
                decimal sum = dstTable[i].GetSum();
                if (Math.Abs(sum) > SumEpsilon) // threshold
                {
                    problems.Add(dstTable[i].Name);
                }
            }
            return problems;
        }

		/// <summary>
		/// Sum the aligned column for the given factor or interaction name and ensure it sums to near-zero.
		/// This effect name can be an interaction expression, like Factor1*Factor2, without spaces.
		/// </summary>
		/// <param name="effect">The factor or interaction for which the target column was aligned.</param>
		/// <param name="response">The name of the response column.</param>
        /// <param name="plainColNames">If true, plain column names are being used.</param>
		/// <param name="dstTable">The table in which the factor and the aligned column reside.</param>
		/// <returns>The name of the aligned column if it fails to sum to near-zero; String.Empty
		/// otherwise.</returns>
		public string SumAlignedColumn(string effect, string response, bool plainColNames, LongFormatDataTable dstTable)
		{
            string efx = (plainColNames && effect.Contains("*")) ? effect.Replace("*", "_x_") : effect;
			string s = String.Format(plainColNames ? "aligned_{0}_for_{1}" : "aligned({0}) for {1}", response, efx);
			LongFormatDataColumn col = dstTable.GetColumn(s);
			if (col != null)
			{
				decimal sum = col.GetSum();
				return (Math.Abs(sum) > SumEpsilon) ? s : String.Empty;
			}
			else
				throw new ArgumentException("Aligned column not found in AlignedRankTransformer.SumAlignedColumn.");
		}

        /// <summary>
        /// Removes any new (computed) columns, which are the aligned or ART columns, except those pertaining
        /// to the given effect name. This effect name can be an interaction expression, like Factor1*Factor2, 
        /// without spaces.
        /// </summary>
        /// <param name="effect">The factor or interaction to except.</param>
        /// <param name="response">The name of the response column.</param>
        /// <param name="plainColNames">If true, plain column names are being used.</param>
        /// <param name="dstTable">The table from which to remove the new columns except for the given effect.</param>
        public void RemoveNewColumnsExceptForEffect(string effect, string response, bool plainColNames, LongFormatDataTable dstTable)
		{
            // build the column names we're exempting from removal
            string efx = (plainColNames && effect.Contains("*")) ? effect.Replace("*", "_x_") : effect;
            string s1 = String.Format(plainColNames ? "aligned_{0}_for_{1}" : "aligned({0}) for {1}", response, efx);
			string s2 = String.Format(plainColNames ? "ART_{0}_for_{1}" : "ART({0}) for {1}", response, efx);

			// column at index 0 is Subjects
			// column at index 1 is a factor
			// column at index 2 could be the original response
			// so start looking at column 3
			for (int i = dstTable.NumColumns - 1; i >= 3; i--)
			{
				string cs = dstTable[i].Name;
				if (cs.StartsWith("aligned(") 
                    || cs.StartsWith("aligned_") 
                    || cs.StartsWith("ART(") 
                    || cs.StartsWith("ART_"))
				{
					if (cs != s1 && cs != s2)
					{
						LongFormatDataColumn col = dstTable.RemoveColumnAt(i);
						Debug.Print("Column '{0}' removed.", col.Name);
					}
				}
			}
		}

		/// <summary>
		/// Takes the existing aligned columns and produces ranks for each of those columns
		/// as new columns. The ranks are averaged in the case of ties.
		/// </summary>
		/// <param name="nFactors">The number of factors that were in the original source table.</param>
		/// <param name="dstTable">The destination table whose aligned columns will be ranked.</param>
		public void ComputeRankedColumns(int nFactors, ref LongFormatDataTable dstTable)
        {
            int startIdx = nFactors + 2; // nFactors + 2 b/c the src table has cols S, X1..Xn, Y
            int count = (int) Math.Pow(2.0, nFactors) - 1;
            for (int i = startIdx; i < startIdx + count; i++)
            {
                List<decimal> ranks = dstTable[i].GetAveragedRanks();
                for (int j = 0; j < ranks.Count; j++)
                {
                    dstTable[i + count].AddRow(ranks[j]);
                }
            }
        }

        /// <summary>
        /// Writes out the aligned-and-ranked data table to a *.csv file.
        /// </summary>
        /// <param name="dstTable">The table of ART-ed data produced by this program.</param>
        /// <param name="outfile">The full path to the output file to be written, including the file name.</param>
        /// <param name="delimiter">The character delimiter used to separate columns.</param>
        public void WriteAlignedRankedTable(LongFormatDataTable dstTable, string outfile, char delimiter)
        {
            StreamWriter writer = new StreamWriter(outfile, false, Encoding.ASCII);

            // write out the column names
            List<string> colnames = dstTable.ColumnNames;
            for (int i = 0; i < colnames.Count; i++)
            {
                if (i != 0)
                    writer.Write(delimiter);
                writer.Write(colnames[i]);
            }
            writer.Write(writer.NewLine);

            // write out the data rows
            for (int i = 0; i < dstTable.NumRows; i++)
            {
                for (int j = 0; j < dstTable.NumColumns; j++)
                {
                    if (j != 0)
                        writer.Write(delimiter);
                    writer.Write(dstTable[j][i]);
                }
                writer.Write(writer.NewLine);
            }
            writer.Close();
        }

        /// <summary>
        /// Tokenizes the given line from a *.csv row using, e.g., commas as delimiters.
        /// </summary>
        /// <param name="line">The character-delimited line parsed from the *.csv file.</param>
        /// <param name="delimiter">The character delimiter used to separate columns.</param>
        /// <returns>A list of string tokens in the order in which they appeared 
        /// on the given line.</returns>
        private List<string> Tokenize(string line, char delimiter)
        {
            List<string> tokens = new List<string>();

            // 1,MI,1,8095,2504,2877,3627.5,2844.5
            // 01234567890123456789012345678901234

            int comma = -1;
            int nextcomma = line.IndexOf(delimiter);
            while (true)
            {
                string token = line.Substring(comma + 1, nextcomma - comma - 1);
                tokens.Add(token.Trim());
                comma = nextcomma;
                if (comma == line.Length)
                    break;
                nextcomma = line.IndexOf(delimiter, comma + 1);
                if (nextcomma == -1) // end of line reached
                    nextcomma = line.Length;
            }

            return tokens;
        }

        /// <summary>
        /// Counts the number of lines in the given file.
        /// </summary>
        /// <param name="file">A full path to an input file.</param>
        /// <returns>The number of lines in the file.</returns>
        /// <remarks>This method is used in debugging only.</remarks>
        private int CountRowsInDataTable(string file)
        {
            int nRows = 0;
            StreamReader reader = new StreamReader(file, Encoding.ASCII);
            while (!reader.EndOfStream)
            {
                reader.ReadLine();
                nRows++;
            }
            reader.Close();
            return nRows;
        }
    }
}
