﻿/**
 * ARTool 2
 * 
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *      
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2010-2024, Jacob O. Wobbrock, Ph.D.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock OR The University
 * of Washington BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;
using WobbrockLib.Controls;

namespace ARTool
{
	/// <summary>
	/// Class for a form that enables the user to specify post hoc pairwise comparisons using the ART-C method.
	/// </summary>
	public partial class ContrastForm : Form
	{
		private readonly int SpaceIndex; // index of Space multifactor separator character
		private readonly string _datafile; // full file path to original data table *.csv
		private readonly string _conExt; // suffix for ART-C file, e.g., ".art-c"
		private readonly LongFormatDataTable _table; // the source data table
		private readonly List<string> _factors; // output selected factors
		private string _outfile; // output file after "OK" is clicked

		/// <summary>
		/// Constructor for the form. 
		/// </summary>
		/// <param name="datafile">The data file from which the contrasts file will be built. For example, a *.csv file path.</param>
		/// <param name="conExt">The contrasts file extension prefix. For example, ".art-c".</param>
		/// <param name="table">The data table object holding the original data.</param>
		public ContrastForm(string datafile, string conExt, bool plainColNames, LongFormatDataTable table)
		{
			InitializeComponent();
			SpaceIndex = ttcboSeparator.Items.Count - 1; // Space is always the bottom item
			
			_datafile = datafile; // "*.csv" file
			_conExt = conExt;     // ".art-c" string
			_outfile = String.Empty;
			_table = table;
			_factors = new List<string>(table.NumFactors);

			// if we are using only plain column names, remove all separator symbols except the 
			// first two (nothing and underscore)...
			if (plainColNames)
			{
				for (int i = ttcboSeparator.Items.Count - 1; i >= 2; i--)
					ttcboSeparator.Items.RemoveAt(i);
			}
		}

		/// <summary>
		/// Gets the factors that were chosen from this dialog.
		/// </summary>
		public List<string> ContrastFactors
		{
			get { return _factors; }
		}

		/// <summary>
		/// Gets the multifactor separator character for merging factor names and levels.
		/// </summary>
		public char MultifactorSeparator
		{
			get 
			{
				int index = ttcboSeparator.SelectedIndex;
				if (index <= 0)
					return '\0';
				else if (index == SpaceIndex)
					return ' ';
				else return ttcboSeparator.Text[0];
			}
		}

		/// <summary>
		/// Gets the output CSV file name that was chosen for ART-C contrasts.
		/// </summary>
		public string Outfile
		{
			get { return _outfile; }
		}

		/// <summary>
		/// Load up the contrast form by setting filenames and factor names and their levels that can be
		/// chosen in a list of checkbox items.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ContrastForm_Load(object sender, EventArgs e)
		{
			txtData.Text = Path.GetFileName(_datafile);
			txtOutput.Text = String.Format("{0}{1}.csv", Path.GetFileNameWithoutExtension(_datafile), _conExt);

			chklstFactors.SuspendLayout();
			List<string> factors = _table.FactorNames;
			for (int i = 0; i < factors.Count; i++)
			{
				LongFormatDataColumn col = _table.GetColumn(factors[i]);
				List<string> levels = col.Levels;

				StringBuilder sb = new StringBuilder(String.Format("{0}: ", factors[i]));
				for (int j = 0; j < levels.Count; j++)
				{
					sb.Append(levels[j]);
					if (j < levels.Count - 1)
						sb.Append(", ");
				}
				chklstFactors.Items.Add(sb.ToString(), false); // e.g., Animal: cat, dog, bird
			}
			chklstFactors.ResumeLayout();
			cmdSelect.Enabled = (chklstFactors.Items.Count > 0);

			txtPrompt.SelectionStart = txtPrompt.TextLength;
			txtPrompt.SelectionLength = 0;

			txtData.SelectionStart = txtData.TextLength;
			txtData.SelectionLength = 0;

			txtOutput.SelectionStart = txtOutput.TextLength;
			txtOutput.SelectionLength = 0;

			Point pt = new Point(this.Owner.Right, this.Owner.Top);
			Rectangle bounds = Screen.GetWorkingArea(pt);
			this.Left = Math.Min(pt.X + this.Width, bounds.Right) - this.Width;
			this.Top = Math.Min(pt.Y + this.Height, bounds.Bottom) - this.Height;
		}

		/// <summary>
		/// When "OK" is clicked, we set the two main exported data items, the factors chosen and the name of
		/// the output CSV file to write for factor contrasts.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cmdOK_Click(object sender, EventArgs e)
		{
			CheckedListBox.CheckedItemCollection items = chklstFactors.CheckedItems;
			for (int i = 0; i < items.Count; i++)
			{
				string s = items[i].ToString();
				string f = s.Substring(0, s.IndexOf(':')); // pull out factor name from, e.g., "Animal: cat, dog, bird"
				_factors.Add(f);
			}
			_outfile = String.Format("{0}\\{1}", Path.GetDirectoryName(_datafile), txtOutput.Text); // update
		}

		/// <summary>
		/// This event handler enables us to check/uncheck clicked upon factor names in the factor list,
		/// and to update the output file name accordingly.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		/// <remarks>This ItemCheck event runs before the list items are actually updated. This hack here forces the items
		/// to be updated manually so that we can process the checked items correctly on the line marked (*), below. See
		/// https://stackoverflow.com/questions/32291324/manage-checkedlistbox-itemcheck-event-to-run-after-an-item-checked-not-before
		/// </remarks>
		private void chklstFactors_ItemCheck(object sender, ItemCheckEventArgs e)
		{
			CheckedListBox thisCtl = (CheckedListBox) sender; // get this checked list box
			thisCtl.ItemCheck -= chklstFactors_ItemCheck;     // remove this event handler so it won't trigger on the next line
			thisCtl.SetItemCheckState(e.Index, e.NewValue);   // manually check or uncheck this item
			thisCtl.ItemCheck += chklstFactors_ItemCheck;     // re-add this event handler

			// update the output file name based on the factors checked
			UpdateOutputFilename(); // (*)

			// enable or disable the separator dropdown
			ttcboSeparator.Enabled = (chklstFactors.CheckedItems.Count > 1);
			lblSeparatorHighlight.Visible = (chklstFactors.CheckedItems.Count > 1);

			// enable or disable the "OK" button
			cmdOK.Enabled = (chklstFactors.CheckedItems.Count > 0);
		}

		/// <summary>
		/// Called when the drop-down containing multiple factor separator characters changes. If multiple factors
		/// are selected, then, optionally, a character can be used to separate their names. The same character will
		/// also then be used to separate their levels' names within the generated file.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ttcboSeparator_SelectedIndexChanged(object sender, EventArgs e)
		{
			UpdateOutputFilename();
		}

		/// <summary>
		/// Updates the output filename in txtOutput based on the factors, if any, that were chosen.
		/// </summary>
		private void UpdateOutputFilename()
		{
			CheckedListBox.CheckedItemCollection items = chklstFactors.CheckedItems;
			if (items.Count > 0)
			{
				StringBuilder sb = new StringBuilder();
				for (int i = 0; i < items.Count; i++)
				{
					string s = items[i].ToString();
					string f = s.Substring(0, s.IndexOf(':')); // pull out factor name
					sb.Append(f);
					if (i < items.Count - 1)
						sb.Append(ttcboSeparator.SelectedIndex == SpaceIndex ? " " : ttcboSeparator.Text);
				}
				txtOutput.Text = String.Format("{0}-{1}{2}.csv", Path.GetFileNameWithoutExtension(_datafile), sb, _conExt);
			}
			else // if no factor names are checked
			{
				txtOutput.Text = String.Format("{0}{1}.csv", Path.GetFileNameWithoutExtension(_datafile), _conExt);
			}
		}

		/// <summary>
		/// Handles when the select all / none button is clicked. If all items are checked, 
		/// clicking this button unchecks them all. But if any are unchecked, then it first
		/// checks them all before another click would uncheck them all.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cmdSelect_Click(object sender, EventArgs e)
		{
			if (chklstFactors.Items.Count == chklstFactors.CheckedIndices.Count) // are all checked?
			{
				for (int i = 0; i < chklstFactors.Items.Count; i++)
				{
					chklstFactors.SetItemChecked(i, false);
				}
			}
			else // if not all are checked, then check them all
			{
				for (int i = 0; i < chklstFactors.Items.Count; i++)
				{
					chklstFactors.SetItemChecked(i, true);
				}
			}
			chklstFactors.ClearSelected(); // clear the blue selection line
		}

		/// <summary>
		/// If a click occurs on the whitespace of the factor list, then clear the selection highlight
		/// from any or all of the factors. Without this handler, there'd be no way to deselect all of
		/// the factors in the factor list.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void chklstFactors_MouseClick(object sender, MouseEventArgs e)
		{
			if (chklstFactors.Items?.Count > 0)
			{
				Rectangle bounds = chklstFactors.GetItemRectangle(chklstFactors.Items.Count - 1);
				if (e.Y > bounds.Bottom)
				{
					chklstFactors.ClearSelected();
				}
			}
		}

		/// <summary>
		/// When the project URL link is clicked in the status bar footer, open the ARTool project page.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lblProjectURL_Click(object sender, EventArgs e)
		{
			string url = lblProjectURL.Tag.ToString();
			Process.Start(url);
		}

		/// <summary>
		/// When the character separator link is clicked upon, give explanatory text so the user knows what this
		/// character separator means.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lnklblSeparator_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			PositionableMessageBox.Show(this,
				new Point(this.Right, this.Top),
				false,
				"This optional character will be used to separate merged\r\n" +
				"contrast factor names and their levels. For example, if\r\n" +
				"one factor is named \"Foo\" and another \"Bar\", then\r\n" +
				"choosing a period (.) will result in a factor named \"Foo.Bar\".\r\n" +
				"The same character is used when merging factors' levels.\r\n" +
				"So if \"Foo\" has a level \"f\" and \"Bar\" has a level \"b\", \r\n" +
				"then the merged contrast factor would be named \"f.b\".",
				"Multifactor Separator Character",
				MessageBoxButtons.OK,
				MessageBoxIcon.Information,
				false);
		}
	}
}
