﻿/**
 * ARTool 2
 * 
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *      
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2010-2021, Jacob O. Wobbrock, Ph.D.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock OR The University
 * of Washington BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using WobbrockLib.Controls;

namespace ARTool
{
	public partial class DataForm : Form
	{
		private List<string> _CSVs;
		private char _delimiter;

		/// <summary>
		/// Constructor for the data form.
		/// </summary>
		/// <param name="CSVs">A list of full paths to CSV data files.</param>
		/// <param name="delimiter">The column delimiter in the CSV data files. This is usually
		/// a comma but could conceivably be something else.</param>
		public DataForm(List<string> CSVs, char delimiter)
		{
			InitializeComponent();
			_CSVs = CSVs;
			_delimiter = delimiter;
		}

		/// <summary>
		/// Handles the load event for the form. Loads up the CSV data files as menu names (with
		/// no menu items) and simulates a click on the first menu, thereby loading it into the data
		/// grid.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void DataForm_Load(object sender, EventArgs e)
		{
			Debug.Assert(_CSVs.Count > 0, "No CSV files were passed to the DataForm.");

			// add filenames and full file path tags to menu strip
			for (int i = 0; i < _CSVs.Count; i++)
			{
				ToolStripMenuItem menu = new ToolStripMenuItem(Path.GetFileNameWithoutExtension(_CSVs[i]), null, MenuFile_OnClick);
				menu.Tag = _CSVs[i]; // full file path
				mnuFiles.Items.Add(menu);
			}

			// simulate a click on the first menu name
			if (mnuFiles.Items.Count > 0)
			{
				mnuFiles.Items[0].PerformClick();
			}
		}

		/// <summary>
		/// Handle when a menu name is clicked to load that file into the data grid.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void MenuFile_OnClick(object sender, EventArgs e)
		{
			// don't reload if the menu that's clicked is alreayd the highlighted one
			ToolStripMenuItem menu = (ToolStripMenuItem) sender;
			if (menu.BackColor == SystemColors.Highlight && menu.ForeColor == SystemColors.HighlightText)
				return;

			// first unhighlight all menu names
			for (int i = 0; i < mnuFiles.Items.Count; i++)
			{
				mnuFiles.Items[i].BackColor = SystemColors.Control;
				mnuFiles.Items[i].ForeColor = SystemColors.ControlText;
			}
			
			// now highlight the selected menu load its CSV into the data grid
			menu.BackColor = SystemColors.Highlight;
			menu.ForeColor = SystemColors.HighlightText;
			LoadCSV(menu.Tag.ToString());
		}

		/// <summary>
		/// Loads a single CSV file into the data grid.
		/// </summary>
		/// <param name="csv">A full path to a CSV file.</param>
		private void LoadCSV(string csv)
		{
			// set the window title to be the shown filename
			this.Text = Path.GetFileName(csv);

			// build up the data table
			DataTable dt = new DataTable();
			string[] lines = File.ReadAllLines(csv, Encoding.UTF8);
			if (lines.Length > 0)
			{
				string[] headings = lines[0].Split(_delimiter);
				for (int i = 0; i < headings.Length; i++) // build columns
				{
					dt.Columns.Add(new DataColumn(headings[i]));
				}
				for (int i = 1; i < lines.Length; i++) // build rows
				{
					string[] values = lines[i].Split(_delimiter);
					DataRow row = dt.NewRow();
					for (int j = 0; j < headings.Length; j++)
					{
						row[j] = values[j];
					}
					dt.Rows.Add(row);
				}
			}

			// set the data source for the data grid view
			if (dt.Rows.Count > 0)
			{
				dgvData.DataSource = dt;
			}

			// set the row headers with row numbers
			for (int i = 0; i < dgvData.Rows?.Count; i++)
			{
				DataGridViewRowHeaderCell rh = dgvData.Rows[i].HeaderCell;
				rh.Value = (i + 1).ToString();
			}
		}

		/// <summary>
		/// When a column header is clicked and rows are sorted, the row header numbers disappear.
		/// So this restores them.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void dgvData_ColumnHeaderMouseClick(object sender, DataGridViewCellMouseEventArgs e)
		{
			// set the row headers with row numbers
			for (int i = 0; i < dgvData.Rows?.Count; i++)
			{
				DataGridViewRowHeaderCell rh = dgvData.Rows[i].HeaderCell;
				rh.Value = (i + 1).ToString();
			}
		}
	}
}
