﻿/**
 * ARTool 2
 * 
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *      
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2010-2021, Jacob O. Wobbrock, Ph.D.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock OR The University
 * of Washington BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;

namespace ARTool
{
    /// <summary>
    /// Represents a single long-format column in a data table. Column values are stored as strings 
    /// although a column can be tested for being numeric, in which case means can be calculated.
    /// </summary>
    public class LongFormatDataColumn
    {
        #region Fields

        private string _name;
        private List<LongFormatDataValue> _values;

        #endregion

        #region Constructor

        /// <summary>
        /// Constructs a new column of data for a data table. The column must
        /// be given a name.
        /// </summary>
        /// <param name="name">The name of the column.</param>
        public LongFormatDataColumn(string name)
        {
            _name = name;
            _values = new List<LongFormatDataValue>(1024);
        }

		/// <summary>
		/// Creates a deep copy of this data column.
		/// </summary>
		/// <returns>A copy of this column.</returns>
		public LongFormatDataColumn Copy()
		{
			LongFormatDataColumn column = new LongFormatDataColumn(_name);
			for (int i = 0; i < _values.Count; i++)
			{
				column.AddRow(_values[i].Value);
			}
			return column;
		}

        /// <summary>
        /// Gets or sets the name of this column.
        /// </summary>
        public string Name
        {
            get
			{
				return _name;
			}
			set
			{
				if (value != String.Empty)
					_name = value;
				else
					throw new ArgumentException("Column name cannot be an empty string in LongFormatDataColumn.Name.");
			}
        }

		/// <summary>
		/// Gets this column as a string, which is just to return its name 
		/// and number of rows.
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			return String.Format("{0} ({1} rows)", _name, _values.Count);
		}

		#endregion

		#region Levels

		/// <summary>
		/// Gets the list of levels contained in this column. This property only
		/// makes sense for subjects and factors, not responses, although nothing
		/// prevents it from being called on responses.
		/// </summary>
		public List<string> Levels
		{
			get
			{
				List<string> seen = new List<string>(_values.Count);
				for (int i = 0; i < _values.Count; i++)
				{
					if (!seen.Contains(_values[i].Value))
					{
						seen.Add(_values[i].Value);
					}
				}
				return seen;
			}
		}

		/// <summary>
		/// Gets the number of distinct levels contained in this column. As this 
		/// method is not presumed useful on numeric outcomes that may use high 
		/// precision, no protection against precision errors is used here.
		/// </summary>
		public int NumLevels // same as number of unique values
		{
			get
			{
				return this.Levels.Count;
			}
		}

		#endregion

		#region Rows

		/// <summary>
		/// Indexer for this column. Gets the row associated with the given 0-based
		/// index. Thus, for row "1", index "0" should be used.
		/// </summary>
		/// <param name="index">The 0-based index for the row to get.</param>
		/// <returns>The string value contained within this column. It may represent 
		/// a numeric type. If the index is out of range, an exception is thrown.</returns>
		public string this[int index] // table row 1 is array index 0, etc.
		{
			get
			{
				return _values[index].Value;
			}
		}

		/// <summary>
		/// Gets the number of rows stored in this column.
		/// </summary>
		public int NumRows
		{
			get { return _values.Count; }
		}

		/// <summary>
		/// Gets a list of 1-based row numbers of all rows that contain the given value.
		/// If no rows contain the value, then an empty but non-null list is returned.
		/// As this method is not presumed useful on numeric outcomes that may use
		/// high precision, no protection against precision errors is used here.
		/// </summary>
		/// <param name="value">The value to search for in the row values.</param>
		/// <returns>A list of 1-based row numbers.</returns>
		public List<int> AllRowsWithLevel(string value) // returns 1-based row numbers
		{
			List<int> rows = new List<int>(_values.Count);
			for (int i = 0; i < _values.Count; i++)
			{
				if (_values[i].Value == value)
				{
					rows.Add(i + 1);
				}
			}
			return rows;
		}

		/// <summary>
		/// Adds a row to the bottom of the column.
		/// </summary>
		/// <param name="item">The item to add. This may be a nominal type with
		/// a character value, or a string representing a numeric value.</param>
		public void AddRow(string item)
        {
            if (item.Trim() == String.Empty)
                throw new ArgumentException("Empty row values are not allowed.");
            else
                _values.Add(new LongFormatDataValue(_values.Count + 1, item));
        }

        /// <summary>
        /// Adds a row with the given decimal value. Any numeric data stored in
        /// a column is regarded as being of the decimal type, as it affords the
        /// highest possible precision.
        /// </summary>
        /// <param name="item"></param>
        public void AddRow(decimal item)
        {
            AddRow(item.ToString());
        }

        #endregion

        #region Calculations

        /// <summary>
        /// Tests this column for whether or not it contains only numeric values.
        /// </summary>
        public bool IsNumeric
        {
            get
            {
				if (_values.Count > 0)
                {
					for (int i = 0; i < _values.Count; i++)
					{
						if (!Decimal.TryParse(_values[0].Value, NumberStyles.Number, null, out decimal m))
							return false; // hit a non-numeric value
					}
					return true; // all values numeric
                }
                return false;
            }
        }

        /// <summary>
        /// Computes the mean over the entire column. Assumes that this column
        /// contains numeric data.
        /// </summary>
        /// <returns>The mean for the entire column.</returns>
        public decimal GetMean()
        {
            return GetMean(null);
        }

        /// <summary>
        /// Computes the mean from column values indicated by the 1-based row numbers
        /// given.
        /// </summary>
        /// <param name="forRows">A list of 1-based row numbers over which to compute the mean.
        /// If null, the mean is computed for the entire column.</param>
        /// <returns>The mean value over the column for the row numbers given.</returns>
        public decimal GetMean(List<int> forRows) // 1-based row numbers
        {
            decimal sum = 0m;
            int count = 0;

            if (forRows == null) // all rows
            {
                for (int i = 0; i < _values.Count; i++)
                {
                    if (Decimal.TryParse(_values[i].Value, NumberStyles.Number, null, out decimal m))
                    {
                        sum += m;
                        count++;
                    }
                    else throw new FormatException(String.Format("Mean calculation failed due to non-numeric value in column '{0}' row {1}.", _name, i + 1));
                }
            }
            else // only selected rows
            {
                for (int i = 0; i < forRows.Count; i++)
                {
                    if (Decimal.TryParse(_values[forRows[i] - 1].Value, NumberStyles.Number, null, out decimal m))
                    {
                        sum += m;
                        count++;
                    }
                    else throw new FormatException(String.Format("Mean calculation failed due to non-numeric value in column '{0}' row {1}.", _name, forRows[i]));
                }
            }
            return sum / count;
        }

        /// <summary>
        /// Computes the sum over the entire column. Assumes that this column
        /// contains numeric data.
        /// </summary>
        /// <returns>The sum for the entire column.</returns>
        public decimal GetSum()
        {
            return GetSum(null);
        }

        /// <summary>
        /// Computes the sum from column values indicated by the 1-based row numbers
        /// given.
        /// </summary>
        /// <param name="forRows">A list of 1-based row numbers over which to compute the mean.
        /// If null, the mean is computed for the entire column.</param>
        /// <returns>The mean value over the column for the row numbers given.</returns>
        public decimal GetSum(List<int> forRows) // 1-based row numbers
        {
            decimal sum = 0m;
            int count = 0;

            if (forRows == null) // all rows
            {
                for (int i = 0; i < _values.Count; i++)
                {
                    if (Decimal.TryParse(_values[i].Value, NumberStyles.Number, null, out decimal m))
                    {
                        sum += m;
                        count++;
                    }
                    else throw new FormatException(String.Format("Sum calculation failed due to non-numeric value in column '{0}' row {1}.", _name, i + 1));
                }
            }
            else // only selected rows
            {
                for (int i = 0; i < forRows.Count; i++)
                {
                    if (Decimal.TryParse(_values[forRows[i] - 1].Value, NumberStyles.Number, null, out decimal m))
                    {
                        sum += m;
                        count++;
                    }
                    else throw new FormatException(String.Format("Sum calculation failed due to non-numeric value in column '{0}' row {1}.", _name, forRows[i]));
                }
            }
            return sum;
        }

        /// <summary>
        /// Assign ranks to the data in this column such that "1" is given to the
        /// smallest value and N is given to the largest value. In the case of ties,
        /// average ranks are used.
        /// </summary>
        /// <remarks>From the JMP 7.0 documentation on averaged ranks: "If a value is unique, 
        /// the averaged rank is the same as the rank. If a value occurs k times, the average 
        /// rank is computed as the sum of the value's ranks divided by k."</remarks>
        public List<decimal> GetAveragedRanks()
        {
            // sort the column by ascending value
            ValueComparer vComp = new ValueComparer();
            _values.Sort(vComp);

            // assign straight ranks
            for (int i = 0; i < _values.Count; i++)
            {
                _values[i].Rank = i + 1m;
            }
            // now look for stretches of value ties
            List<int> ties = new List<int>(_values.Count);
            ties.Add(0); // add initial index; 0 is, of course, tied with itself
            for (int i = 1; i < _values.Count; i++)
            {
                if ((this.IsNumeric && Double.Parse(_values[i-1].Value, NumberStyles.Number, null) == Double.Parse(_values[i].Value,  NumberStyles.Number, null)) // numeric comparison*, or...
                    || (_values[i - 1].Value == _values[i].Value)) // ...string comparison
                {
                    // *parse as doubles to truncate and avoid precision problems at the very end of the decimal
                    ties.Add(i); // add index
                }
                else
                {
                    AssignOneAveragedRank(ties);
                    ties.Clear();
                    ties.Add(i);
                }
            }
            AssignOneAveragedRank(ties);

            // then re-sort by the row value to restore the proper order
            RowComparer rComp = new RowComparer();
            _values.Sort(rComp); // return to ascending row-order
            
            // copy over the averaged ranks for returning
            List<decimal> avgRanks = new List<decimal>(_values.Count);
            for (int r = 0; r < _values.Count; r++)
            {
                avgRanks.Add(_values[r].Rank);
            }
            return avgRanks;
        }

        /// <summary>
        /// Assigns one averaged rank over a range of tied indices.
        /// </summary>
        /// <param name="ties">The contiguous range of indices that have one value 
        /// tied for all of them.</param>
        private void AssignOneAveragedRank(List<int> ties)
        {
            decimal rankSum = 0m;
            for (int j = 0; j < ties.Count; j++)
            {
                rankSum += _values[ties[j]].Rank;
            }
            for (int j = 0; j < ties.Count; j++)
            {
                _values[ties[j]].Rank = rankSum / ties.Count;
            }
        }

        #endregion
    }
}