﻿/**
 * ARTool 2
 * 
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *      
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2010-2021, Jacob O. Wobbrock, Ph.D.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock OR The University
 * of Washington BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using WobbrockLib.Extensions;

namespace ARTool
{
    /// <summary>
    /// Represents a long-format data table with the first column assumed to be Subject, the last
    /// column assumed to be the response, and the intermediate columns assumed to be a factor.
    /// Column values are as strings although a column can be tested for being numeric, in which 
    /// case means can be calculated.
    /// </summary>
    public class LongFormatDataTable
    {
        #region Fields

        private readonly List<LongFormatDataColumn> _columns;

        #endregion

        #region Constructor

        /// <summary>
        /// Constructs a data table that will contain columns of string and numeric data.
        /// </summary>
        public LongFormatDataTable()
        {
            _columns = new List<LongFormatDataColumn>(64);
        }

		/// <summary>
		/// Creates a deep copy of this data table.
		/// </summary>
		/// <returns>A copy of this table.</returns>
		public LongFormatDataTable Copy()
		{
			LongFormatDataTable table = new LongFormatDataTable();
			for (int i = 0; i < _columns.Count; i++)
			{
				table.AddColumn(_columns[i].Copy());
			}
			return table;
		}

		/// <summary>
		/// Gets a string representation of this table.
		/// </summary>
		/// <returns>A string representing this table.</returns>
		public override string ToString()
		{
			return String.Format("Table: {0} columns, {1} rows.", this.NumColumns, this.NumRows);
		}

		#endregion

		#region Counts

		/// <summary>
		/// Gets the number of columns in this data table. By default, the first column 
		/// is Subjects, the last column is the response, and the intermediate
		/// columns are factors.
		/// </summary>
		public int NumColumns
        {
            get { return _columns.Count; }
        }

        /// <summary>
        /// Gets the number of factor columns in this data table. By default, the first 
        /// column is Subject, the last column is the response, and the intermediate
        /// columns are factors.
        /// </summary>
        public int NumFactors
        {
            get { return (_columns.Count > 2) ? _columns.Count - 2 : 0; }
        }

        /// <summary>
        /// Gets the number of rows in this data table. Only the first column is tested,
        /// as all columns are assumed to have the same number of rows.
        /// </summary>
        public int NumRows
        {
            get
            {
                if (_columns.Count > 0)
                {
                    return _columns[0].NumRows;
                }
                else return 0;
            }
        }

        #endregion

        #region Column Names

        /// <summary>
        /// Gets the name of the Subject column. By default, the first column
        /// is Subject, the last column is the response, and the intermediate
        /// columns are factors.
        /// </summary>
        public string SubjectName
        {
            get
            {
                if (_columns.Count > 0)
                {
                    return _columns[0].Name;
                }
                else return String.Empty;
            }
        }

        /// <summary>
        /// Gets the names of the factor columns. By default, the first column
        /// is Subject, the last column is the response, and the intermediate
        /// columns are factors.
        /// </summary>
        public List<string> FactorNames // assumes all columns except the last are X factors (i.v.'s)
        {
            get
            {
                List<string> names = null;
                if (_columns.Count > 2) // first column is Subject and last column is response
                {
                    names = new List<string>(_columns.Count - 2);
                    for (int i = 1; i < _columns.Count - 1; i++) // skip first and last columns
                    {
                        names.Add(_columns[i].Name);
                    }
                }
                return names;
            }
        }

        /// <summary>
        /// Gets the name of the response column. By default, the first column
        /// is Subject, the last column is the response, and the intermediate
        /// columns are factors.
        /// </summary>
        public string ResponseName // assumes the last column in the table is the Y outcome (d.v.)
        {
            get
            {
                if (_columns.Count > 0)
                {
                    return _columns[_columns.Count - 1].Name;
                }
                else return String.Empty;
            }
        }

		/// <summary>
		/// Gets a list of all column names.
		/// </summary>
		public List<string> ColumnNames
		{
			get
			{
				List<string> names = new List<string>(_columns.Count);
				for (int i = 0; i < _columns.Count; i++)
				{
					names.Add(_columns[i].Name);
				}
				return names;
			}
		}

		#endregion

		#region Column Operations

		/// <summary>
		/// Indexer for this table. Gets the column at the 0-based index value.
		/// </summary>
		/// <param name="index">The 0-based index of the column to get.</param>
		/// <returns>The column at the given index. If the index is out of range,
		/// an exception is thrown.</returns>
		public LongFormatDataColumn this[int index]
		{
			get { return _columns[index]; }
		}

		/// <summary>
		/// Adds a column to the "right side" (end) of the data table. The added column 
		/// must have a unique name or an exception is thrown.
		/// </summary>
		/// <param name="col">The column to add.</param>
		public void AddColumn(LongFormatDataColumn col)
        {
            if (col.Name.Trim() == String.Empty)
                throw new ArgumentException("Whitespace column names are not allowed.");
            else if (this.ColumnNames.Contains(col.Name))
                throw new ArgumentException(String.Format("Column name '{0}' is not unique.", col.Name));
            else
                _columns.Add(col);
        }

        /// <summary>
        /// Gets the column with the given name.
        /// </summary>
        /// <param name="name">The name of the column to get.</param>
        /// <returns>The column with the given name, or null.</returns>
        public LongFormatDataColumn GetColumn(string name)
        {
            LongFormatDataColumn col = null;
            for (int i = 0; i < _columns.Count; i++)
            {
                if (_columns[i].Name == name)
                {
                    col = _columns[i];
                    break;
                }
            }
            return col;
        }

		/// <summary>
		/// Gets the first column in this data table, which should be a 
		/// column with subject identifiers.
		/// </summary>
		/// <returns>The first column in this data table, or null if there
		/// are no columns.</returns>
		public LongFormatDataColumn GetFirstColumn()
		{
			LongFormatDataColumn col = null;
			if (_columns?.Count > 0)
				col = _columns[0];
			return col;
		}

		/// <summary>
		/// Gets the last column in this data table, which, if the table has
		/// not been aligned and ranked, will be the response. If it has been
		/// aligned and ranked, it will be an ART-ed column.
		/// </summary>
		/// <returns>The last column in this data table, or null if there
		/// are no columns.</returns>
		public LongFormatDataColumn GetLastColumn()
		{
			LongFormatDataColumn col = null;
			if (_columns?.Count > 0)
				col = _columns[_columns.Count - 1];
			return col;
		}

		/// <summary>
		/// Gets the factor column with the given name. The first and last
		/// columns are not considered, as these are response columns in
		/// an unprocessed data table. The factor columns must lie between these
		/// two bookend columns.
		/// </summary>
		/// <param name="name">The name of the factor column to get.</param>
		/// <returns>The factor column with the given name, or null.</returns>
		public LongFormatDataColumn GetFactorColumn(string name)
		{
			LongFormatDataColumn col = null;
			for (int i = 1; i < _columns.Count - 1; i++)
			{
				if (_columns[i].Name == name)
				{
					col = _columns[i];
					break;
				}
			}
			return col;
		}

        /// <summary>
		/// Find sthe column with the given name and removes it, returning that 
		/// column.
		/// </summary>
		/// <param name="name">The name of the column to remove.</param>
		/// <returns>The column removed, or null if there is no column with the given
		/// name.</returns>
		public LongFormatDataColumn RemoveColumn(string name)
		{
			LongFormatDataColumn col = null;
			for (int i = 0; i < _columns.Count; i++)
			{
				if (_columns[i].Name == name)
				{
					col = _columns[i];
					_columns.RemoveAt(i);
					break;
				}
			}
			return col;
		}

		/// <summary>
		/// Removes the column at the given zero-based index, and returns that column.
		/// </summary>
		/// <param name="index">The zero-based index of the column to remove.</param>
		/// <returns>The removed column, or null if the index is out of range.</returns>
		public LongFormatDataColumn RemoveColumnAt(int index)
		{
			LongFormatDataColumn col = null;
			if (0 <= index && index < _columns.Count)
			{
				col = _columns[index];
				_columns.RemoveAt(index);
			}
			return col;
		}

		/// <summary>
		/// Merges the given factor columns by taking their values and concatenating them, optionally with a separator
        /// character in between for readability. For example, if two factors "X1" and "X2" are merged, the new column 
        /// name will be "X1X2" if no separator character is used, and if, say, a period is used, the name will be
        /// "X1.X2". Similarly, if a level of "X1" is "a" and a level of "X2" is "x", the merged column level will be 
        /// "ax" if no separator character is used, and "a.x" if, say, a period is used.
		/// </summary>
		/// <param name="factors">The names of the factor columns to merge.</param>
        /// <param name="separator">An optional separator character for use between factor names and levels. If no character
        /// is intended to be used, then the null character ('\0') should be passed in.</param>
		/// <param name="removeOriginals">If true, the original columns whose names are given will be removed from this table; 
        /// if false, those original columns will be kept.</param>
		/// <returns>The name of the merged factor column. This table will place the new merged factor just before the original 
        /// response column.</returns>
		public string MergeFactors(List<string> factors, char separator, bool removeOriginals)
		{
			if (factors == null || factors.Count == 0)
				throw new ArgumentException("Invalid factor list passed to LongFormatDataTable.MergeFactors. List is empty.");

			// define the new merged factor name
			StringBuilder sbf = new StringBuilder();
            for (int i = 0; i < factors.Count; i++)
            {
                sbf.Append(factors[i]);
                if (separator != '\0' && i < factors.Count - 1)
                    sbf.Append(separator);
            }

			LongFormatDataColumn mergedColumn = new LongFormatDataColumn(sbf.ToString());

			// gather all of the columns for the factor names given
			List<LongFormatDataColumn> cols = new List<LongFormatDataColumn>(factors.Count);
			for (int k = 0; k < factors.Count; k++)
			{
				LongFormatDataColumn c = this.GetFactorColumn(factors[k]);
				if (c != null)
					cols.Add(c);
				else
					throw new ArgumentException(String.Format("Invalid factor '{0}' passed to LongFormatDataTable.MergeFactors. Factor not found.", factors[k]));
			}

			// create the merged factor levels
			for (int i = 0; i < this.NumRows; i++)
			{
				StringBuilder sbl = new StringBuilder(); // new level via merging
                for (int j = 0; j < cols.Count; j++)
                {
                    sbl.Append(cols[j][i]);
                    if (separator != '\0' && j < cols.Count - 1)
                        sbl.Append(separator);
                }
                mergedColumn.AddRow(sbl.ToString());
			}

			// add the newly merged column just before the response column
			_columns.Insert(_columns.Count - 1, mergedColumn);

			// if desired, remove the original columns that were just merged
			if (removeOriginals)
			{
				for (int f = 0; f < factors.Count; f++)
					this.RemoveColumn(factors[f]);
			}

			return sbf.ToString();
		}

        #endregion

        #region Calculations

        /// <summary>
        /// Gets the subject (S) for the given 1-based row number. By default, the first 
        /// column is Subject, the last column is the response, and the intermediate
        /// columns are factors.
        /// </summary>
        /// <param name="row">The 1-based row number of the subject identifier to get.</param>
        /// <returns>The subject identifier row this row.</returns>
        public string GetSubjectForRow(int row)
        {
            LongFormatDataColumn sCol = _columns[0];
            return sCol[row - 1];
        }

        /// <summary>
        /// Gets the Y response value for the given 1-based row number. This value is
        /// assumed to be numeric. If it is not, an exception is thrown. By default, the 
        /// first  column is Subject, the last column is the response, and the intermediate
        /// columns are factors.
        /// </summary>
        /// <param name="row">The 1-based row number of the response to get.</param>
        /// <returns>A numeric response value.</returns>
        /// <remarks>If the caller intends to compare the decimal value returned by
        /// this method to other decimal values, he might consider first casting it
        /// to a double so as to avoid precision problems at the very end of the decimal.</remarks>
        public decimal GetResponseForRow(int row) // 1-based row number
        {
            LongFormatDataColumn yCol = _columns[_columns.Count - 1];
            if (Decimal.TryParse(yCol[row - 1], NumberStyles.Number, null, out decimal m))
                return m; // Y response
            else
                throw new FormatException(String.Format("Non-numeric response found in column {0} row {1}.", yCol.Name, row));
        }

        /// <summary>
        /// Gets a list of the values of the factors (X's) for the given 1-based row number. Any 
        /// numeric values will be returned as strings. By default, the first column is Subject, 
        /// the last column is the response, and the intermediate columns are factors.
        /// </summary>
        /// <param name="row">The 1-based row number whose values are to be gotten.</param>
        /// <returns>A list of string values or null if two or fewer columns exist in this table.</returns>
        public List<string> GetLevelsOfFactorsForRow(int row) // 1-based row number
        {
            List<string> levels = null;
            if (_columns.Count > 2)
            {
                levels = new List<string>(_columns.Count - 2);
                for (int i = 1; i < _columns.Count - 1; i++) // skip first and last cols
                {
                    levels.Add(_columns[i][row - 1]);
                }
            }
            return levels;
        }

        /// <summary>
        /// Gets the mean of the Y response column for all cells in which the given factor equals
        /// the given level. That is, level 'level' is a level of factor 'factor' in this table.
        /// </summary>
        /// <param name="factor">The factor of interest.</param>
        /// <param name="level">The corresponding level for the given factor.</param>
        /// <returns>The average Y response for the level of the given factor.</returns>
        /// <remarks>If the Y response column, i.e., the last column in the table, is non-numeric, 
        /// an exception is thrown.</remarks>
        public decimal GetMeanForLevelOfFactor(string factor, string level)
        {
            // This method patches in to the more general version that takes N factors and N levels.
            List<string> factors = new List<string>(1);
            factors.Add(factor);
            List<string> levels = new List<string>(1);
            levels.Add(level);
            return GetMeanForLevelsOfFactors(factors, levels);
        }

        /// <summary>
        /// Gets the mean of the Y response column for all cells in which the given factors equal
        /// the given levels. Note that levels[i] is the desired level for factors[i].
        /// </summary>
        /// <param name="factors">The factors of interest. This should contain one or more factor (X) 
        /// names. It may also contain the Subject (S) column name if a subject mean is desired.</param>
        /// <param name="levels">The corresponding level for each factor.</param>
        /// <returns>The average Y response for the levels of the given factors.</returns>
        /// <remarks>If the Y response column, i.e., the last column in the table, is non-numeric, 
        /// an exception is thrown.</remarks>
        public decimal GetMeanForLevelsOfFactors(List<string> factors, List<string> levels)
        {
            List<int> forRows = new List<int>(this.NumRows); // final row set to compute over
            for (int i = 0; i < factors.Count; i++)
            {
                for (int j = 0; j < _columns.Count - 1; j++) // skip outcome column, but include S column
                {
                    if (_columns[j].Name == factors[i])
                    {
                        List<int> rows = _columns[j].AllRowsWithLevel(levels[i]); // not precision-protected
                        forRows = (forRows.Count > 0) ? SetsEx.Intersect<int>(forRows, rows) : forRows = rows;
                        break; // proceed to next factor name
                    }
                }
            }
            LongFormatDataColumn yCol = _columns[_columns.Count - 1];
            if (yCol.IsNumeric)
                return yCol.GetMean(forRows);
            else
                throw new FormatException(String.Format("Unable to compute mean on non-numeric response column '{0}'.", yCol.Name));
        }

        /// <summary>
        /// Computes the grand mean over the last column in the table. If the last
        /// column is not numeric, an exception is thrown. By default, the first 
        /// column is Subject, the last column is the response, and the intermediate
        /// columns are factors.
        /// </summary>
        /// <returns>The mean of the last column in the table.</returns>
        public decimal GetGrandMean() // assumes the last column in the table is an outcome measure and can be averaged
        {
            LongFormatDataColumn yCol = _columns[_columns.Count - 1];
            if (yCol.IsNumeric)
                return yCol.GetMean();
            else
                throw new FormatException(String.Format("Unable to compute mean on non-numeric response column '{0}'.", yCol.Name));
        }

        #endregion
    }
}
