﻿/**
 * ARTool 2
 * 
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *      
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2010-2024, Jacob O. Wobbrock, Ph.D.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock OR The University
 * of Washington BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.Threading;
using System.Text;
using System.IO;
using WobbrockLib.Controls;
using WobbrockLib.Windows;

namespace ARTool
{
    /// <summary>
    /// The MainForm is the main user interface for the Align Rank Transform utility.
    /// </summary>
    public partial class MainForm : Form
    {
		#region Constants 

		private const string kDefaultArtExtension = ".art";
		private const string kDefaultContrastSuffix = "-c";
		private const string kDefaultDelimiter = ",";
        private const string kTabText = "tab";
        private const string kSpaceText = "spc";

		// HTML codes
		private const string kCheckmarkHtml = "&#10003;";
		private const string kHeavyCheckHtml = "&#10004;";
		private const string kExclamationHtml = "&#9888;";
		private const string kDeleteHtml = "&#10008;";

		#endregion

		#region Members

		private readonly AlignedRankTransformer _art;
        private readonly List<string> _msgs;

		#endregion

		#region Start & Stop

		/// <summary>
		/// Constructor for the main user interface form.
		/// </summary>
		public MainForm()
        {
            InitializeComponent();
            _art = new AlignedRankTransformer();
            _msgs = new List<string>(32);
        }

        /// <summary>
        /// Handles the load event for when the MainForm is first shown. Sets the data directory
        /// to be the current directory of the application.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            cboDecimal.Text = ControlPanelEx.DecimalSeparator;
            txtDirectory.Text = Directory.GetCurrentDirectory();
			//Debug.Print(Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location)); // equivalent
			//Debug.Print(Path.GetDirectoryName(Application.ExecutablePath)); // equivalent
			txtDirectory.SelectionStart = txtDirectory.TextLength;
            txtDirectory.SelectionLength = 0;
            this.Show();
            txtDirectory.Focus();
        }

		/// <summary>
		/// Paint a boundary around the web browser output control, since it doesn't paint a nice
		/// boundary for itself.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void MainForm_Paint(object sender, PaintEventArgs e)
		{
			Rectangle rc = webOutput.Bounds;
			rc.X--;
			rc.Y--;
			rc.Width += 2;
			rc.Height += 2;
			e.Graphics.DrawRectangle(Pens.Gray, rc);
		}

		#endregion

		#region About Box

		/// <summary>
		/// Handles when the About hyperlink is clicked and opens the About Box for this application.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lklblAbout_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            AboutForm frm = new AboutForm();
            frm.ShowDialog(this);
        }

		#endregion

		#region Hyperlink Help Text

		/// <summary>
		/// Handles when the label above the directory text box is clicked. Opens a message box
		/// that explains how *.csv files should be formatted.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lklblDirectory_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(this.Right, this.Top),
                false,
                "ARTool requires CSV files to be long-format data tables, which means\r\n" + 
                "they contain only one response per row. The first row should be unique\r\n" + 
                "column names. The first column should be a subject identifier (S), and\r\n" + 
                "the last column should be a numeric response (Y). All columns between\r\n" + 
                "should be factors (X1, X2, X3, etc.). You can use whatever column\r\n" +
				"names you want, provided they have these meanings.",
                "Data Format",
                MessageBoxButtons.OK,
                MessageBoxIcon.Information,
                false);
        }

		/// <summary>
		/// Handles when the help label explaining excluding extensions is clicked.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lklblExclude_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			PositionableMessageBox.Show(this,
				new Point(this.Right, this.Top),
				false,
				"When ARTool creates output CSV files, it adds this extension before\r\n" +
				"the *.csv extension (e.g., *.art.csv). If files for contrast testing are\r\n" +
				"requested, it also adds \"-c\" to this extension (e.g., *.art-c.csv).\r\n" +
				"ARTool then refrains from listing files it has created with these\r\n" + 
				"extensions.",
				"File Extensions",
				MessageBoxButtons.OK,
				MessageBoxIcon.Information,
				false);
		}

		/// <summary>
		/// Handles when the want contrasts help link is clicked. 
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lnklblContrasts_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			chkContrasts.Checked = (PositionableMessageBox.Show(this,
				new Point(this.Right, this.Top),
				false,
				"This option allows you to align-and-rank data for contrast tests.\r\n" +
				"If this box is checked, you will be prompted during the aligning-\r\n" +
				"and-ranking process to indicate factors whose levels you want to\r\n" +
				"directly compare. You then will be given a second CSV output file\r\n" +
				"that will enable you to conduct your desired contrast tests, which\r\n" +
				"typically follow statistically significant main effects or interactions.\r\n" +
				"Would you like to have this box checked?",
				"Want Contrasts?",
				MessageBoxButtons.YesNo,
				MessageBoxIcon.Information,
				false) == DialogResult.Yes);
		}

		/// <summary>
		/// Handles when the decimal separator help link is clicked.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lklblDecimal_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			PositionableMessageBox.Show(this,
				new Point(this.Right, this.Top),
				false,
				"Numbers are parsed according to the Windows Region control panel\r\n" +
				"settings. Of particular importance is the decimal separator. When\r\n" +
				"ARTool starts, this drop-down reflects the control panel setting. If\r\n" +
				"the file you are parsing uses a different decimal separator (e.g., 1,25\r\n" +
				"instead of 1.25), then change this drop-down accordingly. Note that\r\n" +
				"changing this drop-down does not permanently change the Windows\r\n" +
				"Region control panel setting; this change is only temporary.",
				"Decimal Separators",
				MessageBoxButtons.OK,
				MessageBoxIcon.Information,
				false);
		}

		/// <summary>
		/// Handles when the delimiter help label is clicked.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lklblDelimter_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			PositionableMessageBox.Show(this,
				new Point(this.Right, this.Top),
				false,
				"You can define the characters that separate columns here. The\r\n" +
				"first is used when reading in CSV data files, whether for aligning\r\n" +
				"and ranking or just viewing. The second is used when writing out\r\n" +
				"aligned-and-ranked CSV files. The default for each is a comma.",
				"Delimiters",
				MessageBoxButtons.OK,
				MessageBoxIcon.Information,
				false);
		}

		/// <summary>
		/// Handles when the plain text for column names help label is clicked.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void lnklblPlainCols_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			PositionableMessageBox.Show(this,
				new Point(this.Right, this.Top),
				false,
				"By default, the column names written out in your *.art.csv and\r\n" +
				"*.art-c.csv files contain spaces, parentheses, hyphens, and asterisks.\r\n" +
				"Some statistics programs do not accept these characters in their\r\n" +
				"column names. If you select this option, these characters will not\r\n" +
				"be used, replaced instead by only letters and underscores.",
				"Plain Column Names",
				MessageBoxButtons.OK,
				MessageBoxIcon.Information,
				false);
		}

		#endregion

		#region Data Files 

		/// <summary>
		/// Handles when the "..." browse button is clicked next to the text directory. Allows
		/// for a Folder Browser Dialog to be used for selecting the data file directory.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cmdBrowse_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog dlg = new FolderBrowserDialog();
			dlg.Description = "Select a folder containing long-format *.csv data files.";
			dlg.RootFolder = Environment.SpecialFolder.Desktop; // must set for SelectedPath to show
			dlg.SelectedPath = Directory.Exists(txtDirectory.Text) ? txtDirectory.Text : Directory.GetCurrentDirectory();
			dlg.ShowNewFolderButton = false;

			if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                txtDirectory.Text = dlg.SelectedPath;
                txtDirectory.SelectionStart = txtDirectory.TextLength;
                txtDirectory.SelectionLength = 0;
                txtDirectory.Focus();
            }
        }

        /// <summary>
        /// Handles when the directory text field changes, and tries to load any *.csv files from
        /// the directory in the checked listbox. Files that have been ART-ed will also have a *.csv
        /// extension, but we don't want to show those as candidates for ART-ing. Therefore, any files
        /// that have the extension added in the txtExtension text box before the *.csv extension
        /// will be filtered out and not shown.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void txtDirectory_TextChanged(object sender, EventArgs e)
        {
            chklstFiles.SuspendLayout();
            chklstFiles.Items.Clear();

			// enumerate through the files in the directory, pulling out any with CSV extensions that are not excluded
			if (Directory.Exists(txtDirectory.Text))
			{
				string[] files = Directory.GetFiles(txtDirectory.Text, "*.csv");
				string artExt = String.Format("{0}.csv", txtExtension.Text); // main ART file extension, e.g., *.art.csv
				string conExt = String.Format("{0}{1}.csv", txtExtension.Text, kDefaultContrastSuffix); // contrast test file extension, e.g. *.art-c.csv

				for (int i = 0; i < files.Length; i++)
				{
					if (!files[i].EndsWith(artExt) && !files[i].EndsWith(conExt))
					{
						chklstFiles.Items.Add(Path.GetFileNameWithoutExtension(files[i]), false);
					}
				}
			}

            chklstFiles.ResumeLayout();
            cmdSelect.Enabled = (chklstFiles.Items.Count > 0);
			cmdView.Enabled = false;
            cmdAlign.Enabled = false;
			cmdAlign.BackColor = SystemColors.Control;
        }

		/// <summary>
		/// Files that are ART-ed are given an extension (before the real *.csv extension) indicating that
		/// they have been ART-ed. (The default is *.art.csv.) Thus, files with this extension should be
		/// excluded from the checked listbox. When the extension changes, we patch into the handler
		/// for the directory changing, since these work together.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void txtExtension_TextChanged(object sender, EventArgs e)
		{
			// update the contrast test extension label
			txtContrast.Text = String.Format("{0}{1}", txtExtension.Text, kDefaultContrastSuffix);
			txtContrast.SelectionStart = txtContrast.TextLength;
			txtContrast.SelectionLength = 0;

			// now just patch into the handler for when the data directory path has changed
			txtDirectory_TextChanged(sender, e);
		}

		/// <summary>
		/// Ensure that the extension typed into the textbox begins with a period when focus leaves that box.
		/// Also, if the box is empty, ensure that the default extension is there. Also ensure that there's
		/// no whitespace hanging around the extension.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void txtExtension_Leave(object sender, EventArgs e)
		{
			string txt = txtExtension.Text.Trim();
			if (txt == String.Empty)
				txtExtension.Text = kDefaultArtExtension;
			else if (txt[0] != '.')
				txtExtension.Text = txt.Insert(0, ".");
			else
				txtExtension.Text = txt; // ensure trimmed
		}

		#endregion

		#region Widget Events

		/// <summary>
		/// If the main form is resized, the text box that holds the often long directory path
		/// needs to have its contents adjusted so that the path is revealed to be as visible
		/// in full as possible, while favoring keeping the end of the path visible. This 
		/// little hack does the trick.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void MainForm_ResizeEnd(object sender, EventArgs e)
		{
			txtDirectory.SelectionLength = 0;
			txtDirectory.SelectionStart = 0;
			txtDirectory.SelectionStart = txtDirectory.TextLength;
		}

		/// <summary>
		/// As a shortcut, a button for selecting or deselecting all of the checked listbox files
		/// at once is provided. When this button is clicked, if anything is unchecked, all items 
		/// are checked. If all items are checked, all items are unchecked.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cmdSelect_Click(object sender, EventArgs e)
        {
            // if all are checked, uncheck all
            if (chklstFiles.Items.Count == chklstFiles.CheckedIndices.Count)
            {
                for (int i = 0; i < chklstFiles.Items.Count; i++)
                {
                    chklstFiles.SetItemChecked(i, false);
                }
            }
            else // if not all are checked, then check
            {
                for (int i = 0; i < chklstFiles.Items.Count; i++)
                {
                    chklstFiles.SetItemChecked(i, true);
                }
            }
            chklstFiles.ClearSelected(); // clear the blue selection line
        }

        /// <summary>
        /// Handler that enables/disables the align and rank button based on whether any items
        /// are checked or not. Unfortunately, the handler fires <i>before</i> the list box's
        /// checked items are updated, which means that we can't simply examine whether any
        /// items are checked or not. Instead, we have to infer what action just happened and
        /// whether or not any items are checked still or not.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void chklstFiles_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            if (chklstFiles.CheckedIndices.Count == 0 || chklstFiles.CheckedIndices.Count >= 2)
            {
				cmdView.Enabled = true;
                cmdAlign.Enabled = true;
				cmdAlign.BackColor = Color.Honeydew;
            }
            else if (chklstFiles.CheckedIndices.Count == 1)
            {
				cmdView.Enabled = (e.CurrentValue != CheckState.Checked);
                cmdAlign.Enabled = (e.CurrentValue != CheckState.Checked);
				cmdAlign.BackColor = (cmdAlign.Enabled ? Color.Honeydew : SystemColors.Control);
            }
        }

		/// <summary>
		/// When the user clicks in the checklist of files below any filename, if there's a file
		/// selected, it toggles that item's checkbox. So we detect when a click occurs below
		/// any of the checklist's items, and simply clear the selection line. The checkbox
		/// state is left unchanged, which is what we want.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void chklstFiles_MouseClick(object sender, MouseEventArgs e)
		{
			if (chklstFiles.Items?.Count > 0)
			{
				Rectangle bounds = chklstFiles.GetItemRectangle(chklstFiles.Items.Count - 1);
				if (e.Y > bounds.Bottom)
				{
					chklstFiles.ClearSelected();
				}
			}
		}

		/// <summary>
		/// Handles when the View button is clicked and shows any selected CSV files in a grid view.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cmdView_Click(object sender, EventArgs e)
		{
			// gather up the full path files to any selected filenames
			chklstFiles.ClearSelected(); // clear the blue selection line
			CheckedListBox.CheckedItemCollection cic = chklstFiles.CheckedItems;
			List<string> CSVs = new List<string>(cic.Count);
			IEnumerator ii = cic.GetEnumerator();
			while (ii.MoveNext())
			{
				string name = (string) ii.Current; // filename
				string file = String.Format("{0}\\{1}.csv", txtDirectory.Text, name); // full path to input file
				if (File.Exists(file)) // ensure file exists
				{
					CSVs.Add(file); // build up list of full file paths
				}
			}
			char delimiter = GetDelimiter(txtDelimiterIn);
			DataForm frm = new DataForm(CSVs, delimiter);
			frm.ShowDialog(this); // blocking
		}

		/// <summary>
		/// Handles when the decimal separator drop-down is changed. The current thread's 
		/// decimal separator is changed accordingly. This is a control panel setting, but
		/// the setting is not permanently changed; it is only changed for the current 
		/// thread.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cboDecimal_SelectedIndexChanged(object sender, EventArgs e)
        {
            ControlPanelEx.DecimalSeparator = cboDecimal.Text;
        }

        /// <summary>
        /// Handles when input focus leaves the character delimiter text box. If an invalid entry
        /// is left in the text box, this event replaces it with the default delimiter.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void txtDelimiter_Leave(object sender, EventArgs e)
        {
            TextBox t = (TextBox) sender;
            if (t.Text.Length != 1 && t.Text != kTabText && t.Text != kSpaceText)
            {
                t.Text = kDefaultDelimiter;
            }            
        }

        /// <summary>
        /// Handles when the tab or space keys are pressed in the character delimiter
        /// text box. Replaces tab with "tab" and space with "spc" in the box. Regular
        /// characters are handled in the key press event handler.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void txtDelimiter_KeyDown(object sender, KeyEventArgs e)
        {
            TextBox t = (TextBox) sender;
            if (e.KeyCode == Keys.Tab)
            {
                t.Text = kTabText;
                t.SelectionStart = t.TextLength;
				t.SelectionLength = 0;
                e.Handled = true;
            }
            else if (e.KeyCode == Keys.Space)
            {
                t.Text = kSpaceText;
                t.SelectionStart = t.TextLength;
				t.SelectionLength = 0;
				e.Handled = true;
            }
        }

        /// <summary>
        /// Handles when a non-control non-space character is pressed in the character delimiter
        /// text box. Since only one character can be the delimiter, whatever is pressed replaces
        /// whatever was in that text box previously.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void txtDelimiter_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar) && e.KeyChar != ' ')
            {
                TextBox t = (TextBox) sender;
                t.Text = e.KeyChar.ToString();
                t.SelectionStart = t.TextLength;
				t.SelectionLength = 0;
				e.Handled = true;
            }
        }

        /// <summary>
        /// Gets the column character delimiter from the given text box.
        /// </summary>
        /// <param name="t">The text box from which to get the column delimiter.</param>
        /// <returns>The column delimiter.</returns>
        private char GetDelimiter(TextBox t)
        {
			if (t.TextLength == 1)
			{
				return t.Text[0];
			}
			else if (t.Text == kTabText)
			{
				return '\t';
			}
			else if (t.Text == kSpaceText)
			{
				return ' ';
			}
			else
				throw new ApplicationException("Invalid delimiter in TextBox in MainForm.GetDelimiter.");
        }

		#endregion

		#region Aligning and Ranking

		/// <summary>
		/// Flags used to indicate how text should be appended to the HTML output 
		/// that offers feedback during the aligning-and-ranking process.
		/// </summary>
		[Flags]
		private enum AppendHtmlFlags
		{
			None = 0x00,
			Red = 0x01,
			Green = 0x02,
			Bold = 0x04,
			NewlineBefore = 0x08,
			NewlineAfter = 0x10,
			HorizontalRule = 0x20,

			BoldRed = Red | Bold,
			BoldGreen = Green | Bold,

			RedNewlineBefore = Red | NewlineBefore,
			BoldRedNewlineAfter = Red | Bold | NewlineAfter,
			BoldGreenNewlineAfter = Green | Bold | NewlineAfter
		}

		/// <summary>
		/// Handles when the Align and Rank button is pressed. This is the main entry point into
		/// the logic that performs the Aligned Rank Transform procedure on selected files. Also, 
		/// this is where the try/catch logic is placed for all file handling and calculating
		/// operations that take place within it.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cmdAlign_Click(object sender, EventArgs e)
		{
			// clear the blue selection line and output text box before beginning to process new files
			chklstFiles.ClearSelected(); // clear the blue selection line
			if (_msgs.Count > 0)
			{
				_msgs.Clear();
				webOutput.DocumentText = "<HTML/>";
				webOutput.Invalidate();
				Application.DoEvents();
				Thread.Sleep(250);
			}

			CheckedListBox.CheckedItemCollection cic = chklstFiles.CheckedItems;
			AppendMessage(AppendHtmlFlags.None, "Preparing to align and rank {0} file(s)...", cic.Count);
			webOutput.Invalidate();
			Application.DoEvents();
			Thread.Sleep(1000);

			// enumerate through the checked items, which are filenames without their *.csv extensions
			IEnumerator ii = cic.GetEnumerator();
			while (ii.MoveNext())
			{
				string name = (string) ii.Current; // filename
				string file = String.Format("{0}\\{1}.csv", txtDirectory.Text, name); // full path to input file
				if (!File.Exists(file)) // ensure file exists
				{
					AppendMessage(AppendHtmlFlags.Red, "File '{0}' no longer exists. ", name);
					AppendMessage(AppendHtmlFlags.BoldRedNewlineAfter, kDeleteHtml);
				}
				else // *.csv file does exist
				{
					try
					{
						// 1. parse the *.csv file and create a long-format data table.
						AppendMessage(AppendHtmlFlags.HorizontalRule, "Parsing '{0}'. ", name);
						LongFormatDataTable srcTable = _art.ParseDataTable(file, GetDelimiter(txtDelimiterIn));
						AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);
						string outfile = String.Format("{0}{1}.csv", file.Substring(0, file.LastIndexOf('.')), txtExtension.Text);
						AppendMessage(AppendHtmlFlags.None, "Creating '{0}' for ART omnibus tests. ", Path.GetFileName(outfile));
						AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

						// 2. create columns in the destination table and then fill them with aligned data.
						AppendMessage(AppendHtmlFlags.None, "Aligning '{0}' with {1} factor(s) for {2} effect(s). ", name, srcTable.NumFactors, Math.Pow(2.0, srcTable.NumFactors) - 1.0);
						LongFormatDataTable dstTable = _art.CreateNewColumnNames(srcTable, chkPlainCols.Checked);
						_art.ComputeAlignedColumns(srcTable, ref dstTable);
						AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

						// 3. sum each aligned column to ensure it equals ~zero.
						AppendMessage(AppendHtmlFlags.None, "Verifying {0} aligned column(s) in '{1}'. ", Math.Pow(2.0, srcTable.NumFactors) - 1.0, name);
						List<string> problems = _art.SumAlignedColumns(srcTable.NumFactors, dstTable);
						for (int i = 0; i < problems.Count; i++)
							AppendMessage(AppendHtmlFlags.RedNewlineBefore, "Column '{0}' does not sum to zero. {1} ", problems[i], kExclamationHtml);
						AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

						// 4. assign midranks to each of the aligned columns, creating new columns.
						AppendMessage(AppendHtmlFlags.None, "Ranking {0} aligned column(s) in '{1}'. ", Math.Pow(2.0, srcTable.NumFactors) - 1.0, name);
						_art.ComputeRankedColumns(srcTable.NumFactors, ref dstTable);
						AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

						// 5. write out the destination table with the same name as the source table, plus *.art.csv.
						AppendMessage(AppendHtmlFlags.Green, "Writing '{0}' for ART omnibus tests. ", Path.GetFileName(outfile));
						_art.WriteAlignedRankedTable(dstTable, outfile, GetDelimiter(txtDelimiterOut));
						AppendMessage(AppendHtmlFlags.BoldGreenNewlineAfter, kHeavyCheckHtml);

						// 6. if contrasts were requested, prompt for which factors should be involved.
						if (chkContrasts.Checked)
						{
							AppendMessage(AppendHtmlFlags.NewlineBefore, "Prompting for <i>post hoc</i> contrasts... ");
							ContrastForm dlg = new ContrastForm(file, txtContrast.Text, chkPlainCols.Checked, srcTable);
							if (dlg.ShowDialog(this) == DialogResult.OK)
							{
								string confile = dlg.Outfile;
								List<string> factors = dlg.ContrastFactors;
								char separator = dlg.MultifactorSeparator;
								AppendMessage(AppendHtmlFlags.None, "{0} factor(s) selected. ", factors.Count);
								AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

								// 7. create a new data table for contrast testing. merge factors involved in the contrasts  
								//    and drop the originals of those merged. keep factors not involved in contrasts as-is.
								AppendMessage(AppendHtmlFlags.None, "Creating '{0}' for ART-C contrast tests. ", Path.GetFileName(confile));
								LongFormatDataTable conSrcTable = srcTable.Copy();
								string conName = conSrcTable.MergeFactors(factors, separator, true);
								AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

								// 8. align the data for this new contrast factor in a new contrast table.
								AppendMessage(AppendHtmlFlags.None, "Aligning new contrast factor '{0}' from {1} factor(s). ", conName, factors.Count);
								LongFormatDataTable conDstTable = _art.CreateNewColumnNames(conSrcTable, chkPlainCols.Checked);
								_art.ComputeAlignedColumns(conSrcTable, ref conDstTable);
								AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

								// 9. perform the sum-to-zero check on the aligned column.
								AppendMessage(AppendHtmlFlags.None, "Verifying aligned contrast column '{0}'. ", conName);
								string problem = _art.SumAlignedColumn(conName, conSrcTable.ResponseName, chkPlainCols.Checked, conDstTable);
								if (problem != String.Empty)
									AppendMessage(AppendHtmlFlags.RedNewlineBefore, "Column '{0}' does not sum to zero. {1} ", problem, kExclamationHtml);
								AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

								// 10. assign midranks to the aligned data for the ART columns.
								AppendMessage(AppendHtmlFlags.None, "Ranking aligned contrast column '{0}'. ", conName);
								_art.ComputeRankedColumns(conSrcTable.NumFactors, ref conDstTable);
								AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

								// 11. remove extra aligned and ranked columns that are not involved in the contrasts among requested factors.
								string sc = Path.GetFileNameWithoutExtension(confile);
								AppendMessage(AppendHtmlFlags.None, "Finalizing table columns in '{0}'. ", sc.Substring(0, sc.LastIndexOf('.')));
								_art.RemoveNewColumnsExceptForEffect(conName, conSrcTable.ResponseName, chkPlainCols.Checked, conDstTable);
								LongFormatDataColumn last = conDstTable.GetLastColumn();
								last.Name = last.Name.Replace(chkPlainCols.Checked ? "ART_" : "ART(", chkPlainCols.Checked ? "ARTC_": "ART-C(");
								AppendMessage(AppendHtmlFlags.NewlineAfter, kCheckmarkHtml);

								// 12. write out the destination table to the contrast test output file, e.g., *.art-c.csv.
								AppendMessage(AppendHtmlFlags.Green, "Writing '{0}' for ART-C contrast tests. ", Path.GetFileName(confile));
								_art.WriteAlignedRankedTable(conDstTable, confile, GetDelimiter(txtDelimiterOut));
								AppendMessage(AppendHtmlFlags.BoldGreenNewlineAfter, kHeavyCheckHtml);
							}
							else // Cancel was selected
							{
								AppendMessage(AppendHtmlFlags.Red, "No factors selected. ");
								AppendMessage(AppendHtmlFlags.BoldRedNewlineAfter, kDeleteHtml);
							}
						} // end contrast tests
					}
					catch (Exception ex)
					{
						AppendMessage(AppendHtmlFlags.RedNewlineBefore, ex.Message);
						AppendMessage(AppendHtmlFlags.BoldRedNewlineAfter, " {0}", kDeleteHtml); // two spaces intentional
					}
				}
			}
		}

		/// <summary>
		/// Appends an additional message to the HTML output provided by the application. This
		/// output is shown in a Web browser control so that we can easily format the text
		/// and display Unicode characters. We implement this as an accruing set of messages
		/// surrounded by HTML formatting; thus, with each new message, we completely rebuild
		/// and reset the HTML displayed in the Web browser control.
		/// </summary>
		/// <param name="f">Formatting flags for determining how this text should be shown.</param>
		/// <param name="msgFormat">A String.Format()-style formatting string for the message itself.</param>
		/// <param name="args">Optional arguments for the String.Format()-style message.</param>
		private void AppendMessage(AppendHtmlFlags f, string msgFormat, params object[] args)
		{
			// encode the HTML for this particular message
			string m = String.Format("{0}{1}{2}{3}{4}{5}{6}{7}{8}{9}",
				(f & AppendHtmlFlags.HorizontalRule) != 0 ? "<hr/>" : "",
				(f & AppendHtmlFlags.NewlineBefore) != 0 ? "<br/>" : "",
				(f & AppendHtmlFlags.Red) != 0 ? "<span style='color:red'>" : "",
				(f & AppendHtmlFlags.Green) != 0 ? "<span style='color:green'>" : "",
				(f & AppendHtmlFlags.Bold) != 0 ? "<b>" : "",
				String.Format(msgFormat, args),
				(f & AppendHtmlFlags.Bold) != 0 ? "</b>" : "",
				(f & AppendHtmlFlags.Green) != 0 ? "</span>" : "",
				(f & AppendHtmlFlags.Red) != 0 ? "</span>" : "",
				(f & AppendHtmlFlags.NewlineAfter) != 0 ? "<br/>" : "");
			_msgs.Add(m); // add this new message

			// now build all of the extant messages into a full HTML block
			StringBuilder html = new StringBuilder(
				"<HTML><HEAD><STYLE>body{{margin:4px;font-size:9pt;font-family:'Microsoft Sans Serif','Arial','Helvetica';}}</STYLE></HEAD><BODY>"
			);
			for (int i = 0; i < _msgs.Count; i++)
			{
				html.Append(_msgs[i]);
			}
			html.Append("<br/></BODY></HTML>");

			// finally, set all of this as the HTML for the Web browser control
			webOutput.DocumentText = html.ToString();

			// scroll to the bottom of the HTML text so it is in view
			Application.DoEvents();
			webOutput.Document.Window.ScrollTo(0, webOutput.Document.Body.ScrollRectangle.Height);
		}



		#endregion
	}
}
