###
### Higgins.R
###
### Jacob O. Wobbrock, Ph.D.
### The Information School
### University of Washington
### wobbrock@uw.edu
### May 6, 2024
### 

### This R file demonstrates the use of the R version of package ARTool.
### It also demonstrates having Windows ARTool.exe align-and-rank data
### and then here we conduct the relevant ANOVAs. Thus, the results of
### R's ARTool package and Windows ARTool.exe can be compared.


library(plyr)
library(dplyr)
library(car)
library(lme4)
library(lmerTest)
library(multcomp)
library(emmeans)
library(ARTool)


##
## ART with R's package ARTool
##
df1 <- read.csv("Higgins1990-Table5.csv")
df1$Tray = as.factor(df1$Tray)
df1$Moisture = as.factor(df1$Moisture)
df1$Fertilizer = as.factor(df1$Fertilizer)
contrasts(df1$Moisture) <- "contr.sum"
contrasts(df1$Fertilizer) <- "contr.sum"
View(df1)

m1 = art(DryMatter ~ Moisture * Fertilizer + (1|Tray), data=df1)
anova(m1)
#                           F Df Df.res     Pr(>F)    
# Moisture             23.833  3      8 0.00024199 ***
# Fertilizer          122.402  3     24 1.1124e-14 ***
# Moisture:Fertilizer   5.118  9     24 0.00064665 ***


##
## ART-C with R's package ARTool
##
art.con(m1, "Moisture:Fertilizer", adjust="holm") %>%
  summary() %>%  # add significance stars to the output
  mutate(sig. = symnum(p.value, corr=FALSE, na=FALSE,
                       cutpoints = c(0, .001, .01, .05, .10, 1),
                       symbols = c("***", "**", "*", ".", " ")))



##
## ART with Windows ARTool.exe -- manually replicate the above results
##

# We must first run Windows ARTool on Higgins1990-Table5.csv to produce
# this new *.art.csv output file. We check the option for "Output plain
# column names only".
df2 <- read.csv("Higgins1990-Table5.art.csv")
df2$Tray = as.factor(df2$Tray)
df2$Moisture = as.factor(df2$Moisture)
df2$Fertilizer = as.factor(df2$Fertilizer)
contrasts(df2$Moisture) <- "contr.sum"
contrasts(df2$Fertilizer) <- "contr.sum"
View(df2)

# verify
sum(df2$aligned_DryMatter_for_Moisture)              # ~0
sum(df2$aligned_DryMatter_for_Fertilizer)            # ~0
sum(df2$aligned_DryMatter_for_Moisture_x_Fertilizer) # ~0

m2 = lmer(ART_DryMatter_for_Moisture ~ Moisture * Fertilizer + (1|Tray), data=df2) # Moisture
a = Anova(m2, type=3, test.statistic="F")
a[match(c("Moisture"), rownames(a)),]
#                F Df Df.res    Pr(>F)    
# Moisture 23.8326  3      8  0.000242 ***

m3 = lmer(ART_DryMatter_for_Fertilizer ~ Moisture * Fertilizer + (1|Tray), data=df2) # Fertilizer
a = Anova(m3, type=3, test.statistic="F")
a[match(c("Fertilizer"), rownames(a)),]
#                F Df Df.res    Pr(>F)    
# Fertilizer 122.4  3     24 1.112e-14 ***

m4 = lmer(ART_DryMatter_for_Moisture_x_Fertilizer ~ Moisture * Fertilizer + (1|Tray), data=df2) # Moisture:Fertilizer
a = Anova(m4, type=3, test.statistic="F")
a[match(c("Moisture:Fertilizer"), rownames(a)),]
#                         F Df Df.res    Pr(>F)    
# Moisture:Fertilizer 5.118  9     24 0.0006466 ***

# Each result agrees with the art() result, above.


##
## ART-C with Windows ARTool.exe
##

# When we ran Windows ARTool on Higgins1990-Table5.csv, if we checked 
# "Want contrasts", then we got a second dialog in which we could chose
# the factors whose levels we wish to compare. Choosing Moisture and 
# Fertilizer, 
df3 <- read.csv("Higgins1990-Table5-MoistureFertilizer.art-c.csv")
View(df3)
df3$Tray = as.factor(df3$Tray)
df3$MoistureFertilizer = as.factor(df3$MoistureFertilizer)
contrasts(df3$MoistureFertilizer) <- "contr.sum"
levels(df3$MoistureFertilizer) # verify

# verify
sum(df3$aligned_DryMatter_for_Moisture_Fertilizer) # ~0

# run the full factorial model
m5 = lmer(ARTC_DryMatter_for_MoistureFertilizer ~ MoistureFertilizer + (1|Tray), data=df3)
summary(glht(m5, emm(pairwise ~ MoistureFertilizer)), test=adjusted(type="holm"))

